'''
Created on Feb 9, 2020

@author: Dietmar Wolz
'''

import time
    
from fcmaes.astro import MessFull, Messenger, Gtoc1, Cassini1
from fcmaes.optimizer import Optimizer, dtime, logger
from fcmaes.retry import Store, retry, minimize
from fcmaes.testfun import RastriginMean
from fcmaes import cmaes

def messengerFullLoop():    
    while True:    
        problem = MessFull()
        logger.info(problem.name + ' cmaes c++')
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = 40000, 
                       max_evaluations = 50000, useCpp = True)
            
        
def parallel_execution_example(dim, n):
    maxEval = 10000
    popsize = 32
    testfun = RastriginMean(dim, n)
    sdevs = [1]*dim
    
    t0 = time.perf_counter()
  
    ret = cmaes.minimize(testfun.fun, 
                testfun.bounds, max_evaluations=maxEval, popsize=popsize, 
                input_sigma=sdevs, is_parallel=True)         
    print(ret.fun, dtime(t0))
  
    t0 = time.perf_counter()
    ret = cmaes.minimize(testfun.fun, 
                testfun.bounds, max_evaluations=maxEval, popsize=popsize, 
                input_sigma=sdevs, is_parallel=False)         
    print(ret.fun, dtime(t0))
     

def test_all(max_evals = 50000, num_retries = 2000, num = 20):

    problem = Cassini1()
    logger.info(problem.name + ' cmaes c++')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       max_evaluations = max_evals, useCpp = True)
 
    problem = Messenger()
    logger.info(problem.name + ' cmaes c++')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       max_evaluations = max_evals, useCpp = True)
 
    problem = Gtoc1()
    logger.info(problem.name + ' cmaes c++')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       max_evaluations = max_evals, useCpp = True)
 
    problem = MessFull()
    logger.info(problem.name + ' cmaes c++')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       max_evaluations = max_evals, useCpp = True)

    problem = Cassini1()
    logger.info(problem.name + ' cmaes python')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       max_evaluations = max_evals, useCpp = True)
 
    problem = Messenger()
    logger.info(problem.name + ' cmaes python')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       max_evaluations = max_evals, useCpp = False)

    problem = Gtoc1()
    logger.info(problem.name + ' cmaes python')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       max_evaluations = max_evals, useCpp = False)

    problem = MessFull()
    logger.info(problem.name + ' cmaes python')
    for i in range(num):
        ret = minimize(problem.fun, bounds=problem.bounds, num_retries = num_retries, 
                       max_evaluations = max_evals, useCpp = False)
         

    problem = Cassini1()
    logger.info(problem.name + ' dual annealing')
    for i in range(num):
        store = Store(problem.bounds, max_evals)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.dual_annealing, num_retries)
      
    problem = Messenger()
    logger.info(problem.name + ' dual annealing')
    for i in range(num):
        store = Store(problem.bounds, max_evals)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.dual_annealing, num_retries)
  
    problem = Gtoc1()
    logger.info(problem.name + ' dual annealing')
    for i in range(num):
        store = Store(problem.bounds, max_evals)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.dual_annealing, num_retries)
   
    problem = MessFull()
    logger.info(problem.name + ' dual annealing')
    for i in range(num):
        store = Store(problem.bounds, max_evals)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.dual_annealing, num_retries)
      
 
    problem = Cassini1()
    logger.info(problem.name + ' differential evolution')
    for i in range(num):
        store = Store(problem.bounds, max_evals)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.differential_evolution, num_retries)
 
    problem = Messenger()
    logger.info(problem.name + ' differential evolution')
    for i in range(num):
        store = Store(problem.bounds, max_evals)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.differential_evolution, num_retries)
 
    problem = Gtoc1()
    logger.info(problem.name + ' differential evolution')
    for i in range(num):
        store = Store(problem.bounds, max_evals)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.differential_evolution, num_retries)
 
    problem = MessFull()
    logger.info(problem.name + ' differential evolution')
    for i in range(num):
        store = Store(problem.bounds, max_evals)
        optimizer  = Optimizer(store, 0)
        ret = retry(problem.fun, store, optimizer.differential_evolution, num_retries)

def main():

    test_all()
    #parallel_execution_example(27, 1000)

if __name__ == '__main__':
    main()


'''
Created on Dec 8, 2019

@author: Dietmar Wolz
'''

import time
import os
import math
import random
import ctypes as ct
import numpy as np
from random import Random
import multiprocessing as mp
from multiprocessing import Process
from numpy.random import Generator, MT19937, SeedSequence
from scipy.optimize import OptimizeResult, Bounds

from fcmaes.retry import convertBounds
from fcmaes.optimizer import Optimizer, dtime, logger, fitting, seed_random

os.environ['MKL_DEBUG_CPU_TYPE'] = '5'
os.environ['MKL_NUM_THREADS'] = '1'
os.environ['OPENBLAS_NUM_THREADS'] = '1'

def minimize(fun, 
             bounds, 
             value_limit = math.inf,
             num_retries = 5000,
             logger = logger,
             workers = mp.cpu_count(),
             popsize = 31, 
             min_evaluations = 2000, 
             max_evaluations = 50000, 
             evals_step_size = 1000, 
             check_interval = 100,
             useCpp = False,
             stop_fittness = None,
             ):   
    """Minimization of a scalar function of one or more variables using 
    coordinated parallel CMA-ES retry.
     
    Parameters
    ----------
    fun : callable
        The objective function to be minimized.
            ``fun(x, *args) -> float``
        where ``x`` is an 1-D array with shape (n,) and ``args``
        is a tuple of the fixed parameters needed to completely
        specify the function.
    bounds : sequence or `Bounds`, optional
        Bounds on variables for L-BFGS-B, TNC, SLSQP and
        trust-constr methods. There are two ways to specify the bounds:
            1. Instance of the `scipy.Bounds` class.
            2. Sequence of ``(min, max)`` pairs for each element in `x`. None
               is used to specify no bound.
    value_limit : float, optional
        Upper limit for CMA-ES optimized function values to be stored. 
        This limit needs to be carefully set to a value which is seldom
        found by CMA-ES retry to keep the store free of bad runs.
        The crossover offspring of bad parents can
        cause the algorithm to get stuck at local minima.   
    num_retries : int, optional
        Number of CMA-ES retries.    
    logger : logger, optional
        logger for log output of the retry mechanism. If None, logging
        is switched off. Default is a logger which logs both to stdout and
        appends to a file ``optimizer.log``.
    workers : int, optional
        number of parallel processes used. Default is mp.cpu_count()
    popsize = int, optional
        CMA-ES population size used for all CMA-ES runs.
    min_evaluations : int, optional 
        Initial limit of the number of function evaluations.
    max_evaluations : int, optional
        Final limit of the number of function evaluations.
    evals_step_size : int, optional
        Delta the limit of the number of function evaluations is incremented after 
        ``check_interval`` runs
    check_interval : int, optional
        After ``check_interval`` runs the store is sorted and the evaluation limit
        is incremented by ``evals_step_size``
    useCpp : bool, optional
        Flag indicating use of the C++ CMA-ES implementation. Default is `False` - 
        use of the Python CMA-ES implementation
    stop_fittness : float, optional 
         Limit for fitness value. CMA-ES runs terminate if this value is reached. 
    
    Returns
    -------
    res : scipy.OptimizeResult
        The optimization result is represented as an ``OptimizeResult`` object.
        Important attributes are: ``x`` the solution array, 
        ``fun`` the best function value, ``nfev`` the number of function evaluations,
        ``nit`` the number of CMA-ES iterations, ``status`` the stopping critera and
        ``success`` a Boolean flag indicating if the optimizer exited successfully. """

    store = Store(bounds, min_evaluations = min_evaluations, max_evaluations = max_evaluations, 
                  evals_step_size = evals_step_size, check_interval = check_interval, logger = logger)
    optimizer = Optimizer(store, popsize, stop_fittness)
    optimize = optimizer.cma_cpp if useCpp else optimizer.cma_python
    return retry(fun, store, optimize, num_retries, value_limit, workers)

def retry(fun, store, optimize, num_retries, value_limit = math.inf, workers=mp.cpu_count()):
    sg = SeedSequence()
    rgs = [Generator(MT19937(s)) for s in sg.spawn(workers)]
    proc=[Process(target=retry_loop,
            args=(pid, rgs, fun, store, optimize, num_retries, value_limit)) for pid in range(workers)]
    [p.start() for p in proc]
    [p.join() for p in proc]
    store.sort()
    store.dump()
    return OptimizeResult(x=store.get_x(0), fun=store.get_y(0), nfev=store.get_count_evals(), success=True)

def retry_loop(pid, rgs, fun, store, optimize, num_retries, value_limit):
    seed_random() # make sure cpp random generator for this process is initialized properly
    while store.get_count_runs() < num_retries:
        if crossover(fun, store, optimize, rgs[pid]):
            continue
        try:
            dim = len(store.lower)
            sol, y, evals = optimize(fun, None, Bounds(store.lower, store.upper), 
                                     [random.uniform(0.05, 0.1)]*dim, rgs[pid])
            store.add_result(y, sol, store.lower, store.upper, evals, value_limit)
        except Exception as ex:
            continue
 
def crossover(fun, store, optimize, rg):
    if random.random() < 0.5:
        return False
    y0, guess, lower, upper, sdev = store.limits()
    if guess is None:
        return False
    guess = fitting(guess, lower, upper) # take X from lower
    try:       
        sol, y, evals = optimize(fun, guess, Bounds(lower, upper), sdev, rg)
        store.add_result(y, sol, lower, upper, evals, y0) # limit to y0  
    except:
        return False   
    return True  

class Store(object):
    """thread safe storage for optimization retry results; 
    delivers boundary and initial step size vectors for advanced retry crossover operation."""
        
    def __init__(self, 
                 bounds, # bounds of the objective function arguments
                 min_evaluations = 2000, # start with this number of evaluations
                 max_evaluations = 50000, # maximal number of evaluations
                 evals_step_size = 1000, # increase evaluation number by eval_step_size after sorting
                 check_interval = 100, # sort evaluation store after check_interval iterations
                 capacity = 500, # capacity of the evaluation store
                 logger = logger # if None logging is switched off
               ):
        
        self.lower, self.upper = convertBounds(bounds)
        self.logger = logger        
        self.delta = self.upper - self.lower
        self.capacity = capacity
        self.max_evaluations = max_evaluations
        self.evals_step_size = evals_step_size
        self.check_interval = check_interval       
        self.dim = len(self.lower)
        self.random = Random()
        self.t0 = time.perf_counter();
       
        #shared between processes
        self.add_mutex = mp.Lock()    
        self.check_mutex = mp.Lock()                     
        self.xs = mp.RawArray(ct.c_double, capacity * self.dim)
        self.lowers = mp.RawArray(ct.c_double, capacity * self.dim)
        self.uppers = mp.RawArray(ct.c_double, capacity * self.dim)
        self.ys = mp.RawArray(ct.c_double, capacity)                  
        self.max_evals = mp.RawValue(ct.c_int, min_evaluations)
        self.count_evals = mp.RawValue(ct.c_long, 0)   
        self.count_runs = mp.RawValue(ct.c_int, 0) 
        self.num_stored = mp.RawValue(ct.c_int, 0) 
        self.num_sorted = mp.RawValue(ct.c_int, 0)  
        self.best_y = mp.RawValue(ct.c_double, math.inf) 
        self.worst_y = mp.RawValue(ct.c_double, math.inf)  
                                    
    def eval_num(self):
        return self.max_evals.value
                                              
    def limits(self): 
        """guess, boundaries and initial step size for crossover operation."""
        diff_fac = self.random.uniform(0.5, 1.0);
        lim_fac =  self.random.uniform(2.0, 4.0) * diff_fac;
        with self.add_mutex:
            i, j = self.crossover()
            if i < 0:
                return math.inf, None, None, None, None
            x0 = np.asarray(self.get_x(i))
            x1 = np.asarray(self.get_x(j))
            y0 = np.asarray(self.get_y(i))
            
        deltax = np.abs(x1 - x0)
        delta_bound = np.maximum(0.0001, lim_fac * deltax)
        lower = np.maximum(self.lower, x0 - delta_bound)
        upper = np.minimum(self.upper, x0 + delta_bound)
        sdev = np.maximum(0.001, np.minimum(0.5, diff_fac * deltax / self.delta))        
        return y0, x1, lower, upper, sdev
                
    def distance(self, yprev, y, xprev, x): 
        """mean quadratic X difference to neighbor entry."""
        diff = self.worst_y.value - self.best_y.value
        if diff != 0 and (y - yprev) / diff > 0.01: #enough y distance: accept 
            return 1        
        dx = (np.asarray(x) - np.asarray(xprev)) / self.delta
        return math.sqrt(sum(dx*dx)/self.dim)
 
    def replace(self, i, y, xs, lower, upper):
        """replace entry in store."""
        self.set_y(i, y)
        self.set_x(i, xs)
        self.set_lower(i, lower)
        self.set_upper(i, upper)
       
    def crossover(self): # Choose two good entries for recombination
        """indices of store entries to be used for crossover operation."""
        n = self.num_sorted.value
        if n < 2:
            return -1, -1
        lim = self.random.uniform(min(0.1*n, 1.5), 0.2*n)/n
        for c in range(100):
            i1 = -1
            i2 = -1
            for j in range(n):
                if self.random.random() < lim:
                    if i1 < 0:
                        i1 = j
                    else:
                        i2 = j
                        return i1, i2
        return -1, -1
            
    def sort(self): 
        """sorts all store entries, keep only the 90% best to make room for new ones;
        skip entries having similar x values than their neighbors to preserve diversity"""
        ns = self.num_stored.value
        ys = np.asarray(self.ys[:ns])
        yi = ys.argsort()
        sortRuns = []
        yprev = None
        xprev = None
        for i in range(len(yi)):
            y = ys[yi[i]]
            x = self.get_x(yi[i])
            if i == 0 or self.distance(yprev, y, xprev, x) > 0.15:
                sortRuns.append((y, x, self.get_lower(yi[i]), self.get_upper(yi[i])))
                yprev = y
                xprev = x
        numStored = min(len(sortRuns),int(0.9*self.capacity)) # keep 90% best 
        for i in range(numStored):
            self.replace(i, sortRuns[i][0], sortRuns[i][1], sortRuns[i][2], sortRuns[i][3])
        self.num_sorted.value = numStored  
        self.num_stored.value = numStored     
        self.best_y.value = self.get_y(0);
        self.worst_y.value = self.get_y(numStored-1);
        return numStored        
    
    def add_result(self, y, xs, lower, upper, evals, limit=math.inf):
        """registers an optimization result at the store."""
        with self.add_mutex:
            self.incr_count_evals(evals)
            if y < limit:
                if y < self.best_y.value:
                    self.dump()
                    self.best_y.value = y
                if self.num_stored.value >= self.capacity - 1:
                    self.sort()
                ns = self.num_stored.value
                self.num_stored.value = ns + 1
                self.replace(ns, y, xs, lower, upper)
                    
    def get_x(self, pid):
        return self.xs[pid*self.dim:(pid+1)*self.dim]
    
    def get_y(self, pid):
        return self.ys[pid]

    def get_ys(self):
        return self.ys[:self.num_stored.value]
        
    def get_lower(self, pid):
        return self.lowers[pid*self.dim:(pid+1)*self.dim]

    def get_upper(self, pid):
        return self.uppers[pid*self.dim:(pid+1)*self.dim]
    
    def get_count_evals(self):
        return self.count_evals.value
 
    def get_count_runs(self):
        return self.count_runs.value
   
    def set_x(self, pid, xs):
        self.xs[pid*self.dim:(pid+1)*self.dim] = xs[:]
       
    def set_y(self, pid, y):
        self.ys[pid] = y            

    def set_lower(self, pid, lower):
        self.lowers[pid*self.dim:(pid+1)*self.dim] = lower[:]

    def set_upper(self, pid, upper):
        self.uppers[pid*self.dim:(pid+1)*self.dim] = upper[:]

    def incr_count_evals(self, evals):
        """registers the number of evaluations of an optimization run; 
        trigger sorting after check_interval calls. """
        self.count_runs.value += 1
        if self.count_runs.value % self.check_interval == self.check_interval-1:
            if self.max_evals.value < self.max_evaluations:
                self.max_evals.value += self.evals_step_size
            self.sort()
        self.count_evals.value += evals
        
    def dump(self):
        """logs the current status of the store if logger defined."""
        if self.logger is None:
            return
        Ys = self.get_ys()
        vals = []
        for i in range(min(20, len(Ys))):
            vals.append(round(Ys[i],2))     
        dt = dtime(self.t0)            
        message = '{0} {1} {2} {3} {4:.6f} {5:.2f} {6} {7!s} {8!s}'.format(
            dt, int(self.count_evals.value / dt), self.count_runs.value, self.count_evals.value, 
            self.best_y.value, self.worst_y.value, self.num_stored.value, vals, self.get_x(0))
        self.logger.info(message)



'''
Created on Jan 24, 2020

@author: Dietmar Wolz
'''

import sys
import math
import os
import ctypes as ct
from scipy.optimize import Bounds

basepath = os.path.dirname(os.path.abspath(__file__))
if sys.platform.startswith('linux'):
    libgtoplib = ct.cdll.LoadLibrary(basepath + '/lib/libgtoplib.so')    
else:
    libgtoplib = ct.cdll.LoadLibrary(basepath + '/lib/libgtoplib.dll')

class Astrofun(object):
    """Provides access to ESAs GTOP optimization test functions."""
    def __init__(self, name, fun_c, lower, upper):    
        self.name = name 
        self.fun_c = fun_c
        self.fun_c.argtypes = [ct.c_int, ct.POINTER(ct.c_double)] 
        self.fun_c.restype = ct.c_double           
        self.fun = python_fun(self.fun_c)
        self.bounds = Bounds(lower, upper)

class MessFull(object):
    """ see https://www.esa.int/gsp/ACT/projects/gtop/messenger_full.html """
    def __init__(self):    
        Astrofun.__init__(self, 'messenger full', libgtoplib.messengerfullC, 
                           [1900.0, 3.0,    0.0, 0.0,  100.0, 100.0, 100.0, 100.0, 100.0, 100.0,  0.01, 0.01, 0.01, 0.01, 0.01, 0.01,  1.1, 1.1, 1.05, 1.05, 1.05,  -math.pi, -math.pi, -math.pi, -math.pi, -math.pi],
                           [2200.0, 4.05, 1.0, 1.0,  500.0, 500.0, 500.0, 500.0, 500.0, 550.0,  0.99, 0.99, 0.99, 0.99, 0.99, 0.99,  6.0,   6.0,    6.0,    6.0,    6.0,  math.pi,  math.pi,  math.pi,  math.pi,  math.pi]
        )
     
class Messenger(object):
    """ see https://www.esa.int/gsp/ACT/projects/gtop/messenger_reduced.html """

    def __init__(self):    
        Astrofun.__init__(self, 'messenger reduced', libgtoplib.messengerC, 
                           [1000.,1.,0.,0.,200.,30.,30.,30.,0.01,0.01,0.01,0.01,1.1,1.1,1.1,-math.pi,-math.pi,-math.pi],
                           [4000.,5.,1.,1.,400.,400.,400.,400.,0.99,0.99,0.99,0.99,6,6,6,math.pi,math.pi,math.pi]      
        )
    
class Gtoc1(object):
    """ see https://www.esa.int/gsp/ACT/projects/gtop/gtoc1.html """
    
    def __init__(self):    
        Astrofun.__init__(self, 'GTOC1', libgtoplib.gtoc1C, 
                           [3000.,14.,14.,14.,14.,100.,366.,300.],
                           [10000.,2000.,2000.,2000.,2000.,9000.,9000.,9000.]       
       )

class Cassini1(object):
    """ see https://www.esa.int/gsp/ACT/projects/gtop/cassini1.html """
    
    def __init__(self):    
        Astrofun.__init__(self, 'Cassini1', libgtoplib.cassini1C, 
                           [-1000.,30.,100.,30.,400.,1000.],
                           [0.,400.,470.,400.,2000.,6000.]       
        )
    
def python_fun(cfun):
    return lambda x : call_c(cfun, x)

def call_c(cfun, x):
    n = len(x)
    array_type = ct.c_double * n   
    try:
        val = float(cfun(n, array_type(*x)))
    except:
        val = sys.float_info.max
    return val



'''
Created on Jan 24, 2020

@author: Dietmar Wolz
'''

import sys
import os
import ctypes as ct
import numpy as np
from numpy.random import MT19937, Generator
from scipy.optimize import OptimizeResult

from fcmaes.cmaes import checkBounds

os.environ['MKL_DEBUG_CPU_TYPE'] = '5'

def is_terminate_false(runid, iterations, val):
    return False 

def minimize(fun, 
             bounds=None, 
             x0=None, 
             input_sigma = 1.0, 
             popsize = 32, 
             max_evaluations = 100000, 
             max_iterations = 100000,  
             accuracy = 1.0, 
             stop_fittness = None, 
             is_terminate = None, 
             rg = Generator(MT19937()),
             runid=0):   
    """Minimization of a scalar function of one or more variables using a 
    C++ CMA-ES implementation called via ctypes.
     
    Parameters
    ----------
    fun : callable
        The objective function to be minimized.
            ``fun(x, *args) -> float``
        where ``x`` is an 1-D array with shape (n,) and ``args``
        is a tuple of the fixed parameters needed to completely
        specify the function.
    bounds : sequence or `Bounds`, optional
        Bounds on variables for L-BFGS-B, TNC, SLSQP and
        trust-constr methods. There are two ways to specify the bounds:
            1. Instance of the `scipy.Bounds` class.
            2. Sequence of ``(min, max)`` pairs for each element in `x`. None
               is used to specify no bound.
    x0 : ndarray, shape (n,)
        Initial guess. Array of real elements of size (n,),
        where 'n' is the number of independent variables.  
    input_sigma : ndarray, shape (n,) or scalar
        Initial step size for each dimension.
    popsize = int, optional
        CMA-ES population size.
    max_evaluations : int, optional
        Forced termination after ``max_evaluations`` function evaluations.
    max_iterations : int, optional
        Forced termination after ``max_iterations`` iterations.
    accuracy : float, optional
        values > 1.0 reduce the accuracy.
    stop_fittness : float, optional 
         Limit for fitness value. If reached minimize terminates.
    is_terminate : callable, optional
        Callback to be used if the caller of minimize wants to 
        decide when to terminate. 
    rg = numpy.random.Generator, optional
        Random generator for creating random guesses.
    runid : int, optional
        id used by the is_terminate callback to identify the CMA-ES run. 
            
    Returns
    -------
    res : scipy.OptimizeResult
        The optimization result is represented as an ``OptimizeResult`` object.
        Important attributes are: ``x`` the solution array, 
        ``fun`` the best function value, ``nfev`` the number of function evaluations,
        ``nit`` the number of CMA-ES iterations, ``status`` the stopping critera and
        ``success`` a Boolean flag indicating if the optimizer exited successfully. """
    
    if not sys.platform.startswith('linux'):
        raise Exception("CMAES C++ variant currently only supported on Linux")
    lower, upper, guess = checkBounds(bounds, x0, rg)   
    if lower is None:
        lower = []
        upper = []
    mu = int(popsize/2)
    n = guess.size   
    if np.ndim(input_sigma) == 0:
        input_sigma = [input_sigma] * n
    if stop_fittness is None:
        stop_fittness = np.nan   
    if is_terminate is None:    
        is_terminate=is_terminate_false
        use_terminate = False 
    else:
        use_terminate = True 
    array_type = ct.c_double * n   
    c_callback = call_back_type(c_func(fun))
    c_is_terminate = is_terminate_type(is_terminate)
    try:
        res = optimizeACMA_C(runid, c_callback, n, array_type(*guess), array_type(*lower), array_type(*upper), 
                           array_type(*input_sigma), max_iterations, max_evaluations, stop_fittness, mu, 
                           popsize, accuracy, use_terminate, c_is_terminate)
        x = np.array(np.fromiter(res, dtype=np.float64, count=n))
        val = res[n]
        evals = int(res[n+1])
        iterations = int(res[n+2])
        stop = int(res[n+3])
        freemem(res)
        return OptimizeResult(x=x, fun=val, nfev=evals, nit=iterations, status=stop, success=True)
    except Exception:
        return OptimizeResult(x=None, fun=sys.float_info.max, nfev=0, nit=0, status=-1, success=False)

def c_func(fun):
    """Convert an objective function for serial execution for cmaescpp.
    
    Parameters
    ----------
    fun : objective function mapping a list of float arguments to a float value

    Returns
    -------
    out : function
        A function suitable as ctypes based argument for cmaescpp.minimize."""
 
    return lambda n, x: fun([x[i] for i in range(n)])
  
if sys.platform.startswith('linux'):
    basepath = os.path.dirname(os.path.abspath(__file__))
    libcmalib = ct.cdll.LoadLibrary(basepath + '/lib/libacmalib.so')    
    call_back_type = ct.CFUNCTYPE(ct.c_double, ct.c_int, ct.POINTER(ct.c_double))  
    is_terminate_type = ct.CFUNCTYPE(ct.c_bool, ct.c_long, ct.c_int, ct.c_double)      
    optimizeACMA_C = libcmalib.optimizeACMA_C
    optimizeACMA_C.argtypes = [ct.c_long, call_back_type, ct.c_int, \
                ct.POINTER(ct.c_double), ct.POINTER(ct.c_double), ct.POINTER(ct.c_double), \
                ct.POINTER(ct.c_double), ct.c_int, ct.c_int, ct.c_double, ct.c_int, ct.c_int, \
                ct.c_double, ct.c_bool, is_terminate_type]
    
    optimizeACMA_C.restype = ct.POINTER(ct.c_double)         
    freemem = libcmalib.free_mem
    freemem.argtypes = [ct.POINTER(ct.c_double)]
    seed = libcmalib.seed
    seed.argtypes = [ct.c_int]
    seed_random = libcmalib.seedRandom
    seed_random.argtypes = []
 

